/**
* Automatic_region_growing_3D18C.java
* Created on 17 february 2007, 15:40
* e-mail: mgforero@yahoo.es
*
* Function: Non recursive region growing algorithm. The seeds of the growing
* regions are obtained from a binarized or labelled image of type 16.
*
* This plugin do not create a new image, and the result in charged in the seed
* image. If the seed image is not labelled, it is assumed just one seed by
* object, so more seeds cannot be assigned to the same region. In this case,
* you must be sure of having just one pixel by every seed.
* A threshold is established, so a pixel is added to a region if the difference
* between the intensity of the pixel and the mean of the pixels of the region
* is lower than the threshold.
*
* Copyright (c) 2007 by Manuel Guillermo Forero-Vargas
*
* This plugin is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 
* as published by the Free Software Foundation.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this plugin; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
import ij.*;
import ij.gui.*;
import ij.plugin.*;
import ij.process.*;

public class Automatic_region_growing_3D18C implements PlugIn
{
    static boolean labelledSeeds=false;

    public void run(String arg)
    {
        int a,depth,depthminus1;
        int threshold=10;

        if (arg.equals("about"))
        {
            showAbout();
            return;
        }
        //----------------------------------------------------------------------
        String[] titles;
        int[] wList=WindowManager.getIDList();
        if (wList==null||wList.length<2)
        {
            IJ.showMessage("Region growing","There must be at least two windows open");
            return;
        }
        titles=new String[wList.length];
        for (int i=0;i<wList.length;i++)
            titles[i]=WindowManager.getImage(wList[i]).getTitle();
        //----------------------------------------------------------------------
        int IndexIm,IndexSeed;
        GenericDialog gd=new GenericDialog("Region growing");
        gd.addChoice("Seed:",titles,titles[0]);
        gd.addChoice("Image:",titles,titles[1]);
        gd.addCheckbox("Labelled seeds",labelledSeeds);
        gd.addNumericField("Threshold:",threshold,0);
        gd.showDialog();
        if (gd.wasCanceled())
            return;
        IndexSeed=gd.getNextChoiceIndex();
        IndexIm=gd.getNextChoiceIndex();
        labelledSeeds=gd.getNextBoolean();
        threshold=(int)gd.getNextNumber();
        //----------------------------------------------------------------------
        ImagePlus impSeed=WindowManager.getImage(wList[IndexSeed]);
        if (impSeed.getBitDepth()!=16)
        {
            IJ.showMessage("Region growing","The type of the seed image must be 16");
            return;
        }
        ImagePlus impIm=WindowManager.getImage(wList[IndexIm]);
        if (impIm.getBitDepth()!=8)
        {
            IJ.showMessage("Region growing","The type of the mask image must be 8");
            return;
        }
        depth=impSeed.getStackSize();
        if (depth!=impIm.getStackSize())
        {
            IJ.showMessage("Region growing","The number of images of both stacks must be the same");
            return;
        }
        depthminus1=depth-1;
        //----------------------------------------------------------------------
        //Initialisation
        //----------------------------------------------------------------------
        final int height=impSeed.getHeight();
        final int width=impSeed.getWidth();
        if ((height!=impIm.getHeight())||(width!=impIm.getWidth()))
        {
            IJ.showMessage("Region growing","The size of the images of both stacks must be the same");
            return;
        }
        //----------------------------------------------------------------------
        int labels=0,seedsNumber=0,x,y,z;
        byte[][][] pixels=new byte[depth][width][height];
        ImageStack stackSeed=impSeed.getStack();
        ImageStack stackIm=impIm.getStack();
        short[][][] seedPixels=new short[depth][width][height];
        byte[] pixelsO;
        short[] seedPixelsO;
        //----------------------------------------------------------------------
        //Does a copy of the original and the seed stacks and
        //Get the number of seeds
        if (labelledSeeds)
        {
            for (z=0;z<depth;z++)
            {
                pixelsO=(byte[])stackIm.getProcessor(z+1).getPixels();
                seedPixelsO=(short[])stackSeed.getProcessor(z+1).getPixels();
                 for(y=0;y<height;y++)
                {
                    a=y*width;
                    for(x=0;x<width;x++)
                    {
                        pixels[z][x][y]=pixelsO[a+x];//original
                        seedPixels[z][x][y]=seedPixelsO[a+x];//original
                    }
                }
            }
            StackStatistics stats=new StackStatistics(impSeed);
            labels=(int)stats.histMax+1;
        }
        else
        {
        //----------------------------------------------------------------------
        //Does a copy of the seed stack 
            for (z=0;z<depth;z++)
            {
                pixelsO=(byte[])stackIm.getProcessor(z+1).getPixels();
                seedPixelsO=(short[])stackSeed.getProcessor(z+1).getPixels();
                for(y=0;y<height;y++)
                {
                    a=y*width;
                    for(x=0;x<width;x++)
                    {
                        pixels[z][x][y]=pixelsO[a+x];//original
                        if ((seedPixels[z][x][y]=seedPixelsO[a+x])!=0)//original
                            seedsNumber++;
                    }
                }
            }
            labels=seedsNumber+1;
        }
        //----------------------------------------------------------------------
        double[] mean=new double[labels];//Average of the pixels' intensities belonging to the region
        //Note:Label 0 not employed
        int[] sum=new int[labels];//Sum of the pixels intensities belonging to the region
        int[] pixelsNumber=new int[labels];//Number of pixels of the region
        //------------------------------------------------------------------
        Queue queue=new Queue(256);//256 grey levels
        short label=1;
        if (labelledSeeds)
        {
            for (z=0;z<depth;z++)
                for(y=0;y<height;y++)
                    for(x=0;x<width;x++)
                        if (seedPixels[z][x][y]!=0)
                        {
                            //--------------------------------------------------
                            //Inserts the seed in the queue
                            //--------------------------------------------------
                            label=seedPixels[z][x][y];//Label
                            queue.push(new Pointv3D(x,y,z),pixels[z][x][y]&0xff);
                            sum[label]+=pixels[z][x][y]&0xff;//A region can have several seeds
                            mean[label]=(double)sum[label]/++pixelsNumber[label];
                        }
        }
        else
            for (z=0;z<depth;z++)
                for(y=0;y<height;y++)
                    for(x=0;x<width;x++)
                        if (seedPixels[z][x][y]!=0)
                        {
                            //--------------------------------------------------
                            //Inserts the seed in the queue
                            //--------------------------------------------------
                            seedPixels[z][x][y]=label;//Label
                            queue.push(new Pointv3D(x,y,z),pixels[z][x][y]&0xff);
                            sum[label]=pixels[z][x][y]&0xff;//Just one seed by region
                            mean[label]=(double)sum[label];
                            pixelsNumber[label++]=1;
                        }
        queue.updatePriority();
        //------------------------------------------------------------------
        int i,iMax,q,step=0,X=0,Y=0,Z=0;
        Pointv3D p;
        while(queue.notEmpty())
        {
            if ((++step%10000)==0)
                IJ.showStatus("Reconstructing: "+step);
            //--------------------------------------------------------------
            //Starting from the seeds does the region growing
            //--------------------------------------------------------------
            p=((Pointv3D)queue.pop());
            x=p.x;//x ordinate
            y=p.y;//y ordinate
            z=p.z;//z ordinate
            q=pixels[z][x][y]&0xff;
            label=seedPixels[z][x][y];
            //--------------------------------------------------------------
            //Revision of neighbours
            i=z==0?5:0;//Region growing in the first slice
            iMax=(z==depthminus1)?13:18;//Region growing in the last slice
            for (;i<iMax;i++)
            {
                switch (i)
                {
                    case 0://(x+1,y,z-1)
                        X=x+1;
                        Y=y;
                        Z=z-1;
                        break;
                    case 1://(x,y,z-1)
                        X--;
                        break;
                    case 2://(x,y-1,z-1)
                        Y--;
                        break;
                    case 3://(x-1,y,z-1)
                        X--;
                        Y++;
                        break;
                    case 4://(x,y+1,z-1)
                        X++;
                        Y--;
                        break;
                    case 5://(x+1,y-1,z)
                        X=x+1;
                        Y=y-1;
                        Z=z;
                        break;
                    case 6://(x,y-1,z)
                        X--;
                        break;
                    case 7://(x-1,y-1,z)
                        X--;
                        break;
                    case 8://(x-1,y,z)
                        Y++;
                        break;
                    case 9://(x-1,y+1,z)
                        Y++;
                        break;
                    case 10://(x,y+1,z)
                        X++;
                        break;
                    case 11://(x+1,y+1,z)
                        X++;
                        break;
                    case 12://(x+1,y,z)
                        Y--;
                        break;
                    case 13://(x+1,y,z+1)
                        Z++;
                        break;
                    case 14://(x,y,z+1)
                        X--;
                        break;
                    case 15://(x,y-1,z+1)
                        Y--;
                        break;
                    case 16://(x-1,y,z+1)
                        X--;
                        Y++;
                        break;
                    case 17://(x,y+1,z+1)
                        X++;
                        Y++;
                        break;
                }
                if(X>=0&&X<width&&Y>=0&&Y<height)
                    if (seedPixels[Z][X][Y]==0&&(Math.abs((pixels[Z][X][Y]&0xff)-mean[label])<=threshold))
                {
                    queue.push(new Pointv3D(X,Y,Z),q);
                    seedPixels[Z][X][Y]=label;
                    sum[label]+=pixels[Z][X][Y]&0xff;
                    mean[label]=(double)sum[label]/++pixelsNumber[label];
                }
            }
        }
        //----------------------------------------------------------------------
        //Charge the result
        for (z=0;z<depth;z++)
        {
            seedPixelsO=(short[])stackSeed.getProcessor(z+1).getPixels();
            for(y=0;y<height;y++)
            {
                a=y*width;
                for(x=0;x<width;x++)
                    seedPixelsO[a+x]=seedPixels[z][x][y];//original
            }
        }
        impSeed.show();
        IJ.run("Enhance Contrast", "saturated=0.5");
        impSeed.updateAndDraw();
    }

    void showAbout()
    {
        IJ.showMessage("About Automatic_region_growing_3D18C...",
        "This plug-in applies an iterative region growing algorithm to a stack of 8-bit images.");
    }
}
