/**
 * Fast_hybrid_dome_3D.java
 * Created on 3rd August 2007, 12:05
 *
 * Fast 3D dome algorithm.
 * Function: Localises the domes in a stack by using the Vicent's hybrid method.
 * Based on the Fast_hybrid_grey_scale_reconstruction_3D_8bits plugin.
 *
 * This plugin do not create a new image.
 *
 * Copyright (c) 2007 by Manuel Guillermo Forero-Vargas
 * e-mail: mgforero@yahoo.es
 *
 * This plugin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this plugin; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
import ij.*;
import ij.gui.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import java.util.*;//Employed to use the LinkedList class

public class Fast_hybrid_dome_3D implements PlugInFilter
{
    protected  ImagePlus imp;
    protected static int domeheight=20;

    public int setup(String arg, ImagePlus imp)
    {
        if (arg.equals("about"))
        {
            showAbout();
            return DONE;
        }
        GenericDialog gd=new GenericDialog("Dome 3D");
        gd.addNumericField ("Height",domeheight,0);
        gd.showDialog();
        if (gd.wasCanceled())
            return DONE;
        domeheight=(int)gd.getNextNumber();
        this.imp=imp;
        return DOES_8G+DOES_8C+NO_CHANGES+STACK_REQUIRED;
    }

    public void run(ImageProcessor ip)
    {
        int a,b,c,d,i,iMax,max,offset,offsetz,q;
        int x,X=0,y,Y=0,z,Z=0;
        final int height=ip.getHeight();
        final int width=ip.getWidth();
        final int size=height*width;

        //----------------------------------------------------------------------
        final int depth=imp.getStackSize();
        final int depthminus1=depth-1;
        //----------------------------------------------------------------------
        // Initialisation
        //----------------------------------------------------------------------
        final int sizeS=size*depth;
        int[] originalPixels=new int[sizeS];//Original image
        int[] darkPixels=new int[sizeS];//Dark image
        byte[] resultPixels;
        //----------------------------------------------------------------------
        //Does a copy of the seed and mask stacks
        ImageStack stack=imp.getStack();
        for (z=0;z<depth;z++)
        {
            b=z*size;
            resultPixels=(byte[])stack.getProcessor(z+1).getPixels();
            for(a=0;a<size;a++)
            {
                c=b+a;
                originalPixels[c]=resultPixels[a]&0xff;//original
                darkPixels[c]=originalPixels[c]-domeheight;//darkened
            }
        }
        //----------------------------------------------------------------------
        //Image reconstruction
        //----------------------------------------------------------------------
        LinkedList fifo=new LinkedList();
        //Grey scale dilation of the dark image
        for (z=0;z<depth;z++)
        {
            IJ.showStatus("Fast hybrid dome 3D:First step Iteration "+z+"/"+depth+"...");
            IJ.showProgress(z,depth);
            b=z*size;
            for (y=0;y<height;y++)
            {
                a=b+y*width;
                for (x=0;x<width;x++)
                {
                    c=a+x;
                    i=z==0?9:0;//Limits the region growing in the first slice
                    for (max=0;i<14;i++)
                    {
                        switch (i)
                        {
                            case 0://(x-1,y+1,z-1)
                                X=x-1;
                                Y=y+1;
                                Z=z-1;
                                break;
                            case 1://(x,y+1,z-1)
                                X++;
                                break;
                            case 2://(x+1,y+1,z-1)
                                X++;
                                break;
                            case 3://(x+1,y,z-1)
                                Y--;
                                break;
                            case 4://(x,y,z-1)
                                X--;
                                break;
                            case 5://(x-1,y,z-1)
                                X--;
                                break;
                            case 6://(x-1,y-1,z-1)
                                Y--;
                                break;
                            case 7://(x,y-1,z-1)
                                X++;
                                break;
                            case 8://(x+1,y-1,z-1)
                                X++;
                                break;
                            case 9://(x+1,y-1,z)
                                X=x+1;
                                Y=y-1;
                                Z=z;
                                break;
                            case 10://(x,y-1,z)
                                X--;
                                break;
                            case 11://(x-1,y-1,z)
                                X--;
                                break;
                            case 12://(x-1,y,z)
                                Y++;
                                break;
                            case 13://(x,y,z)
                                X++;
                                break;
                        }
                        if(X>=0&&X<width&&Y>=0&&Y<height&&(q=darkPixels[Z*size+Y*width+X])>max)
                            max=q;
                    }
                    darkPixels[c]=max<originalPixels[c]?max:originalPixels[c];
                }
            }
        }
        for (z=depthminus1;z>=0;z--)
        {
            IJ.showStatus("Fast hybrid dome 3D:Second step Iteration "+z+"/"+depth+"...");
            IJ.showProgress (z,depth);
            b=z*size;
            for (y=height-1;y>=0;y--)
            {
                a=b+y*width;
                for (x=width-1;x>=0;x--)//Gets the pixels inside the window
                {
                    c=a+x;
                    iMax=z==depthminus1?5:14;//Limits the region growing in the last slice
                    for (i=max=0;i<iMax;i++)
                    {
                        switch (i)
                        {
                            case 0://(x,y,z)
                                X=x;
                                Y=y;
                                Z=z;
                                break;
                            case 1://(x+1,y,z)
                                X++;
                                break;
                            case 2://(x+1,y+1,z)
                                Y++;
                                break;
                            case 3://(x,y+1,z)
                                X--;
                                break;
                            case 4://(x-1,y+1,z)
                                X--;
                                break;
                            case 5://(x-1,y+1,z+1)
                                Z++;
                                break;
                            case 6://(x,y+1,z+1)
                                X++;
                                break;
                            case 7://(x+1,y+1,z+1)
                                X++;
                                break;
                            case 8://(x+1,y,z+1)
                                Y--;
                                break;
                            case 9://(x,y,z+1)
                                X--;
                                break;
                            case 10://(x-1,y,z+1)
                                X--;
                                break;
                            case 11://(x-1,y-1,z+1)
                                Y--;
                                break;
                            case 12://(x,y-1,z+1)
                                X++;
                                break;
                            case 13://(x+1,y-1,z+1)
                                X++;
                                break;
                        }
                        if(X>=0&&X<width&&Y>=0&&Y<height&&(q=darkPixels[Z*size+Y*width+X])>max)
                            max=q;
                    }
                    darkPixels[c]=max<originalPixels[c]?max:originalPixels[c];
                    for (i=1;i<iMax;i++)
                    {
                        switch (i)
                        {
                            case 1://(x+1,y,z)
                                X=x+1;
                                Y=y;
                                Z=z;
                                break;
                            case 2://(x+1,y+1,z)
                                Y++;
                                break;
                            case 3://(x,y+1,z)
                                X--;
                                break;
                            case 4://(x-1,y+1,z)
                                X--;
                                break;
                            case 5://(x-1,y+1,z+1)
                                Z++;
                                break;
                            case 6://(x,y+1,z+1)
                                X++;
                                break;
                            case 7://(x+1,y+1,z+1)
                                X++;
                                break;
                            case 8://(x+1,y,z+1)
                                Y--;
                                break;
                            case 9://(x,y,z+1)
                                X--;
                                break;
                            case 10://(x-1,y,z+1)
                                X--;
                                break;
                            case 11://(x-1,y-1,z+1)
                                Y--;
                                break;
                            case 12://(x,y-1,z+1)
                                X++;
                                break;
                            case 13://(x+1,y-1,z+1)
                                X++;
                                break;
                        }
                        if(X>=0&&X<width&&Y>=0&&Y<height)
                        {
                            d=Z*size+Y*width+X;
                            q=darkPixels[d];
                            if(q<darkPixels[c]&&q<originalPixels[d])
                                fifo.add(new Pointv3D(x,y,z));
                        }
                    }
                }
            }
        }
        int step=0;
        Pointv3D p;
        while (!fifo.isEmpty())
        {
            if ((++step%10000)==0)
                IJ.showStatus("Fast hybrid dome 3D:Thirth step "+step);
            //p=(Pointv3D)fifo.pollFirst();//JAVA 1.6
            p=(Pointv3D)fifo.removeFirst();//JAVA 1.4
            x=(int)p.getX();//x ordinate
            y=(int)p.getY();//y ordinate
            z=(int)p.getZ();
            b=z*size+y*width+x;
            for (Z=z-1;Z<=z+1;Z++)//Sweeps pixels inside subwindow
                if (Z>=0&&Z<depth)
                {
                    offsetz=Z*size;
                    for (Y=y-1;Y<=y+1;Y++)//Sweeps pixels inside subwindow
                        if (Y>=0&&Y<height)
                        {
                            offset=offsetz+Y*width;
                            for (X=x-1;X<=x+1;X++)
                                if (X>=0&&X<width)
                                {
                                    c=offset+X;
                                    if (darkPixels[c]<darkPixels[b]&&darkPixels[c]!=originalPixels[c])
                                    {
                                        darkPixels[c]=darkPixels[b]<originalPixels[c]?darkPixels[b]:originalPixels[c];
                                        fifo.addLast(new Pointv3D(X,Y,Z));
                                    }
                                }
                        }
                }
        }
        //----------------------------------------------------------------------
        for (z=0;z<depth;z++)
        {
            IJ.showStatus("Fast hybrid dome 3D:Fourth step Iteration "+z+"/"+depth+"...");
            IJ.showProgress (z,depth);
            b=z*size;
            resultPixels=(byte[])stack.getProcessor(z+1).getPixels();
            for(a=0;a<size;a++)
                resultPixels[a]=(byte)(originalPixels[b+a]-darkPixels[b+a]);
        }
        //----------------------------------------------------------------------
        imp.show();
        imp.updateAndDraw();
        IJ.showStatus("Fast dome 3D: Ready");
     }

    void showAbout()
    {
        IJ.showMessage("About Fast hybrid dome 3D...",
        "This plug-in gets the 3D domes of 8-bit stacks.");
    }
}
