/**
 * IFT_Watershed.java
 * Created on 23th February 2009, 14:40
 * This plug-in filter obtain the watershed of the input stack, based on the
 * labelled seeds, included in the seeded stack.
 *
 * This plugin works in 2D and 3D and does not create a new stack.
 *
 * This plugin is based on the plugin IFT_
 *
 * Copyright (c) 2009 by Manuel Guillermo Forero-Vargas
 * e-mail: mgforero@yahoo.es
 *
 * This plugin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this plugin; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.process.ImageProcessor;
import ij.plugin.filter.PlugInFilter;
import ij.gui.GenericDialog;
import ij.WindowManager;

public class IFT_Watershed implements PlugInFilter
{
    private static final int BYTE=0,SHORT=1;
    private ImagePlus imp;
    private int w,h,d;
    private int wh;
    private  boolean[] flag;

    public int setup(String arg, ImagePlus imp)
    {
        if (arg.equals("about"))
        {
            showAbout();
            return DONE;
        }
        this.imp=imp;
        return DOES_8G+DOES_8C;
    }

    public void run(ImageProcessor ip)
    {
        short[] seedPixels;
        byte[][] pixels;
        short[] result;
        HQueue queue;
        int[] C;
        int a,b,whd,z;
        int typeSeed=4;

        int[] wList=WindowManager.getIDList();
        String[] titles=new String[wList.length];
        for(int i=0;i<wList.length;i++)
            titles[i]=WindowManager.getImage(wList[i]).getTitle();
        GenericDialog gd=new GenericDialog("IFT Watershed");
        gd.addChoice("Seeds",titles,titles[0]);
        gd.showDialog();
        if (gd.wasCanceled())
            return;
        int IndexSeed=gd.getNextChoiceIndex();
        //----------------------------------------------------------------------
        ImagePlus impSeed=WindowManager.getImage(wList[IndexSeed]);
        //----------------------------------------------------------------------
        int bitDepth=impSeed.getBitDepth();
        switch (bitDepth)
        {
            case 8:
                typeSeed=BYTE;
                break;
            case 16:
                typeSeed=SHORT;
                break;
            default:
                break;
        }
        if (typeSeed>1)
        {
            IJ.error("8-bit or 16-bit stacks required");
            return;
        }
        //----------------------------------------------------------------------
        w=imp.getWidth();
        h=imp.getHeight();
        if ((h!=impSeed.getHeight())||(w!=impSeed.getWidth()))
        {
            IJ.error("The size of the images of both stacks must be the same");
            return;
        }
        wh=w*h;
        d=imp.getStackSize();
        if (d!=impSeed.getStackSize())
        {
            IJ.error("The number of images of both stacks must be the same");
            return;
        }
        whd=wh*d;
        //----------------------------------------------------------------------
        //Initialisation of the IFT
        pixels=new byte[d][];
        for(z=0;z<d;z++)//Charge the stack
            pixels[z]=(byte[])imp.getStack().getProcessor(z+1).getPixels();
        C=new int[whd];//Cost matrix
        flag=new boolean[whd];
        result=new short[whd];
        for(int i=0;i<whd;i++)
            C[i]=255;
        queue=new HQueue();
        ImageStack stackSeed=impSeed.getStack();
        for(z=0;z<impSeed.getStackSize();z++)
        {
            if (typeSeed==BYTE)
                seedPixels=(short[])stackSeed.getProcessor(z+1).convertToShort(false).getPixels();
            else
                seedPixels=(short[])stackSeed.getPixels(z+1);
            a=z*wh;
            for(int y=0;y<h;y++)
            {
                int c=y*w;
                for(int x=0;x<w;x++)
                {
                    int i=c+x;
                    if(seedPixels[i]!=0)
                    {
                        int index=a+i;
                        C[index]=pixels[z][i]&0xff;//0;
                        result[index]=seedPixels[i];
                        queue.push(index,C[index]);
                    }
                }
            }
        }
        //----------------------------------------------------------------------
        //Propagation
        IntArray nei=new IntArray(6);
        while(queue.notEmpty())
        {
            int q=queue.pop();
            flag[q]=true;
            nei=getNeighbours(q,nei);
            for(int i=0;i<nei.size();i++)
            {
                int p=nei.get(i);
                int pCost=C[p];
                int m=Math.max(C[q],(int)(pixels[p/wh][p%wh]&0xff));
                if(m<C[p])
                {
                    C[p]=m;
                    result[p/wh*wh+p%wh]=result[q/wh*wh+q%wh];
                    queue.removeFromBucket(p,pCost);
                    queue.push(p,C[p]);
                }
            }
        }
        //----------------------------------------------------------------------
        for(z=0;z<d;z++)
        {
            b=z*wh;
            if (typeSeed==BYTE)
            {
                byte[] seedBytePixels=(byte[])stackSeed.getProcessor(z+1).getPixels();
                for(a=0;a<wh;a++)
                    seedBytePixels[a]=(byte)result[b+a];
            }
            else
            {
                seedPixels=(short[])stackSeed.getPixels(z+1);
                System.arraycopy(result,b,seedPixels,0,wh);
            }
        }
        impSeed.show();
        impSeed.updateAndDraw();
        IJ.showStatus("IFT Watershed: Ready");
    }

    public IntArray getNeighbours(int index,IntArray l)
    {
        l.removeAll();
        int z=index/(wh);
        int s=index%(wh);
        int x=s%w,y=s/w;
        if(z>1&&!flag[(z-1)*wh+s])
            l.add((z-1)*wh+s);
        if(z<d-1&&!flag[(z+1)*wh+s])
            l.add((z+1)*wh+s);
        if(x>1&&!flag[z*wh+s-1])
            l.add(z*wh+s-1);
        if(x<w-1&&!flag[z*wh+s+1])
            l.add(z*wh+s+1);
        if(y>1&&!flag[z*wh+(y-1)*w+x])
            l.add(z*wh+(y-1)*w+x);
        if(y<h-1&&!flag[z*wh+(y+1)*w+x])
            l.add(z*wh+(y+1)*w+x);
        return l;
    }

    void showAbout()
    {
        IJ.showMessage("About IFT_Watershed...",
        "This plug-in filter calculates the 3d IFT watershed of 8-bit stacks.");
    }

    private static class HQueue
    {
        private IntArray[] arr;
        private int nbElements;

        public HQueue()
        {
            arr=new IntArray[256];
            nbElements=0;
        }

        public void push(int position,int cost)
        {
            if(arr[cost]==null)
                arr[cost]=new IntArray(1000);
            arr[cost].add(position);
            nbElements++;
        }

        public boolean notEmpty()
        {
            return nbElements==0?false:true;
        }

        public int pop()
        {
            for(int i=0;i<arr.length;i++)
                if(arr[i]!=null&&!arr[i].isEmpty())
                {
                    int r=arr[i].size()-1;
                    int ret=arr[i].get(r);
                    arr[i].removeIndex(r);
                    nbElements--;
                    return ret;
                }
            return -1;
        }

        public void removeFromBucket(int position,int cost)
        {
            if(arr[cost]!=null)
            {
                arr[cost].removeValue(position);
                nbElements--;
            }
        }
    }

    private static class IntArray
    {
        private int[] array;
        private int size=0;
        private int initCap;

        public IntArray(int cap)
        {
            initCap=cap;
            array=new int[cap];
        }

        public int size()
        {
            return size;
        }

        public boolean isEmpty()
        {
            return size==0;
        }

        public int get(int i)
        {
            return i>=0&&i<size?array[i]:-1;
        }

        public void add(int n)
        {
            if(size==array.length)
            {
                int additional=initCap/5;
                additional=additional==0?1:additional;
                int[] tmp=array;
                array=new int[array.length+additional];
                System.arraycopy(tmp,0,array,0,tmp.length);
            }
            array[size]=n;
            size++;
        }

        public void removeIndex(int i)
        {
            if(i<0||i>=size)
                return;
            size--;
            for(int k=i;k<size;k++)
                array[k]=array[k+1];
        }

        public void removeValue(int q)
        {
            int k=0;
            for(;k<size&&array[k]!=q;k++);
            if(k==size)
                return;
            size--;
            for(;k<size;k++)
                array[k]=array[k+1];
        }

        public void removeAll()
        {
            size=0;
        }
    }
}