/*
* M_Track4.java
* Created on July 2010
*
* This is a modification of the MTrack2 algorithm, developed by Nico Stuurman
* 
* This plugin considers only one object in the stack, therefore there is only
* one track to follow.
*
* Developed by Manuel Guillermo Forero-Vargas
* e-mail: mgforero@yahoo.es
*/

import ij.plugin.filter.PlugInFilter;
import java.awt.Color;
import java.util.*;
import ij.*;
import ij.gui.*;
import ij.text.*;
import ij.process.*;
import ij.plugin.filter.Info;
import ij.plugin.filter.ParticleAnalyzer;
import ij.measure.*;
 
/**
	Uses ImageJ's particle analyzer to track the movement of
	multiple objects through a stack. 
	Based on the Object Tracker plugin filter by Wayne Rasband

	Based on Multitracker,but should be quite a bit more intelligent
	Nico Stuurman,Vale Lab,UCSF/HHMI,May,June 2003
*/

public class M_Track4 implements PlugInFilter,Measurements
{
    protected   ImageStack stack;
    ImagePlus	imp;

    static double   fps=17;//Frame per second. This value is employed if this value cannot be find in the headfile
    static float    maxVelocity=50;
    static int      minSize=90;
    static int      maxSize=1000;
    static int      minTrackLength=20;
    static boolean  bShowPositions=false;
    static boolean  bShowPaths=false;
    static boolean  bShowPathLengths=false;

    public class particle
    {
        float	x;
        float	y;
        float   angle;
        int     frame;
        int     trackNr=0;
        boolean flag=false;

        public double distance(particle p)
        {
            return Math.hypot(this.x-p.x,this.y-p.y);
        }
    }

    public int setup(String arg,ImagePlus imp)
    {
        if (imp==null)
        {
            IJ.noImage();
            return DONE;
        }
        //----------------------------------------------------------------------
        //Creates a dialog box, allowing the user to enter the requested
        //number of standard deviations.
        GenericDialog gd=new GenericDialog("Object Tracker");
        gd.addNumericField("Frames per second:",fps,6,10,"");
        gd.addNumericField("Minimum_object size:",minSize,0,6,"pixels");
        gd.addNumericField("Maximum_object size:",maxSize,0,6,"pixels");
        gd.addNumericField("Maximum_Velocity:",maxVelocity,0);
        gd.addNumericField("Minimum_track length:",minTrackLength,0,6,"frames");
        gd.addCheckbox("Display Path Lengths",bShowPathLengths);
        gd.addCheckbox("Show_Positions",bShowPositions);
        gd.addCheckbox("Show_Paths",bShowPaths);
        gd.showDialog();
        if (gd.wasCanceled())
            return DONE;
        fps=(double)gd.getNextNumber();
        minSize=(int)gd.getNextNumber();
        maxSize=(int)gd.getNextNumber();
        maxVelocity=(float)gd.getNextNumber();
        minTrackLength=(int)gd.getNextNumber();
        bShowPathLengths=gd.getNextBoolean();
        bShowPositions=gd.getNextBoolean();
        bShowPaths=gd.getNextBoolean();
        //----------------------------------------------------------------------
        stack=imp.getStack();
        this.imp=imp;
        return DOES_8G+STACK_REQUIRED+NO_CHANGES;
    }

    public void run(ImageProcessor ip)
    {
        int nFrames=stack.getSize();
        int trackCount=0;

        //Create results table. Keeps the results for each frame
        ResultsTable rt=new ResultsTable();
        //Create storage for particle positions
        List[] theParticles=new ArrayList[nFrames];
        //Record particle positions for each frame in an ArrayList
        for (int frame=0;frame<nFrames;frame++)
        {
            IJ.showProgress((double)frame/nFrames);
            theParticles[frame]=new ArrayList();//Keep the particles of whole frames
            rt.reset();
            ParticleAnalyzer pa=new ParticleAnalyzer(0,CENTROID|ELLIPSE,rt,minSize,maxSize);
            pa.setHideOutputImage(true);
            pa.analyze(imp,stack.getProcessor(frame+1));
            float[] sxRes=rt.getColumn(ResultsTable.X_CENTROID);
            if (sxRes!=null)
            {
                float[] syRes=rt.getColumn(ResultsTable.Y_CENTROID);
                float[] sAngle=rt.getColumn(ResultsTable.ANGLE);
                for (int iPart=0;iPart<sxRes.length;iPart++)
                {
                    particle aParticle=new particle();
                    aParticle.x=sxRes[iPart];
                    aParticle.y=syRes[iPart];
                    aParticle.angle=sAngle[iPart];
                    aParticle.frame=frame;
                    theParticles[frame].add(aParticle);
                }
            }
        }
        //Assemble tracks out of the particle lists
        //Also record to which track a particle belongs in ArrayLists
        List theTracks=new ArrayList();
        for (int frame=0;frame<nFrames;frame++)
        {
            IJ.showProgress((double)frame/nFrames);
            for (ListIterator j=theParticles[frame].listIterator();j.hasNext();)
            {
                particle aParticle=(particle)j.next();
                if (aParticle.trackNr==0)
                {
                    //This must be the beginning of a new track
                    List aTrack=new ArrayList();
                    aParticle.trackNr=++trackCount;
                    aTrack.add((particle)aParticle);
                    //Search in next frames for more particles to be added to track
                    particle oldParticle=aParticle;
                    for (int iF=frame+1;iF<nFrames;iF++)
                    {
                        particle newParticle=null;
                        for (ListIterator jF=theParticles[iF].listIterator();jF.hasNext();)
                        {
                            particle testParticle=(particle)jF.next();
                            double distance=testParticle.distance(oldParticle);
                            //record a particle when it is within the search radius,and when it had not yet been claimed by another track
                            if(distance<maxVelocity)
                            {
                                if (testParticle.trackNr==0)//If this particle has not been tracked yet...
                                {
                                    //if we had not found a particle before,it is easy
                                    if (newParticle==null)
                                        newParticle=testParticle;
                                    else
                                    {
                                        //if we had one before,we'll take this one if it is closer.  In any case,flag these particles
                                        newParticle.flag=true;
                                        testParticle.flag=true;
                                        if (distance<newParticle.distance(oldParticle))
                                            newParticle=testParticle;
                                    }
                                }
                                else
                                {
                                    int i;
                                //This particle is already in another track but could have been part of this one.
                                // We have a number of choices here:
                                // 1. Sort out to which track this particle really belongs (but how?)
                                // 2. Stop this track
                                // 3. Stop this track, and also delete the remainder of the other one
                                // 4. Stop this track and flag this particle (Used in Mtrack2). Comment
                                //some of the following instructions to follow this choice.
                                //5.The track follows the closest particle. The other track is stopped and flag both particles:
                                    testParticle.flag=true;
                                    //------------------------------------------
                                    //Comment from this point if you want to use
                                    //option 4.
                                    List bTrack=(ArrayList)theTracks.get(testParticle.trackNr-1);
                                    for (i=0;i<bTrack.size();i++)
                                        if (((particle)bTrack.get(i)).frame==iF)
                                            break;
                                    double distance1=((particle)bTrack.get(i-1)).distance((particle)bTrack.get(i));
                                    if (distance<distance1)//If the new particle is closer
                                    {
                                        int a=bTrack.size();
                                        for (int b=i;i<a;i++)
                                        {
                                            particle tmpParticle=(particle)bTrack.get(b);
                                            tmpParticle.trackNr=trackCount;
                                            aTrack.add(tmpParticle);
                                            bTrack.remove(b);
                                            iF=nFrames;//End the big for
                                        }
                                        break;
                                    }
                                    //------------------------------------------
                                    //Comment until this point if you want to use
                                    //option 4.
                                }
                            }
                        }
                        if (newParticle!=null)
                        {
                            newParticle.trackNr=trackCount;
                            aTrack.add((particle)newParticle);
                        }
                        else
                            break;
                        oldParticle=newParticle;
                    }
                    theTracks.add(aTrack);
                }
            }
        }
        //------------------------------------------------------------------
        TextWindow resultsWindow=new TextWindow("Track Results","",400,250);
        TextPanel textPanel=resultsWindow.getTextPanel();
        //------------------------------------------------------------------
        //Get the frames per second of the stack
        Info inf=new Info();
        String info=inf.getImageInfo(imp,ip);
        int cc=info.indexOf("Frame rate:");
        if (cc!=-1)
        {
            info=info.substring(cc);
            cc=1+info.indexOf(":");
            int cd=info.indexOf(" fps");
            info=info.substring(cc,cd);
            fps=Double.valueOf(info).doubleValue();
        }
        //------------------------------------------------------------------
        //Create the column headings based on the number of tracks with length
        //greater than minTrackLength since the number of tracks can be larger
        //than can be accomodated by Excell,we deliver the tracks in chunks of maxColumns
        // As a side-effect,this makes the code quite complicated
        String str="Frame";
        int[] validTracks=new int[theTracks.size()+1];
        int track=1;
        ArrayList tmpArray;
        for (ListIterator iT=theTracks.listIterator();iT.hasNext();)
            if ((tmpArray=(ArrayList)iT.next()).size()>=minTrackLength)
            {
                if (track==1)
                    str+="\tX\tY\tAngle\tFlag\tInstant distance\tInstant speed";
                track++;
                particle tmpParticle=(particle)tmpArray.get(0);
                validTracks[tmpParticle.trackNr]=tmpParticle.frame;//Mark the initial frame of the track
            }
            else
                validTracks[((particle)tmpArray.get(0)).trackNr]=-1;//Mark the invalid tracks

        //IJ.setColumnHeadings(str);
        textPanel.setColumnHeadings(str);
        resultsWindow.show();

        String flag;
        double x=0,y=0;
        for (int i=0;i<nFrames;i++)//Write the position of particles in frames
        {
            str=""+(i+1);//str=Position of particle in frame
            for (int a=1;a<validTracks.length;a++)
            {
                if (validTracks[a]!=-1)//If it is a valid track...
                {
                    List bTrack=(ArrayList)theTracks.get(a-1);
                    if (i>=validTracks[a]&&i<validTracks[a]+bTrack.size())
                    {
                        particle aPar=(particle)bTrack.get(i-validTracks[a]);
                        flag=aPar.flag?"*":"";
                        if (i==validTracks[a])
                        {
                            str+="\t"+aPar.x+"\t"+aPar.y+"\t"+aPar.angle+"\t"+flag+"\t \t";
                        }
                        else
                        {
                            float instantDistance=(float)Math.sqrt((aPar.x-x)*(aPar.x-x)+(aPar.y-y)*(aPar.y-y));
                            float instantSpeed=(float)(instantDistance*fps);
                            str+="\t"+aPar.x+"\t"+aPar.y+"\t"+aPar.angle+"\t"+flag+"\t"+instantDistance+"\t"+instantSpeed;
                        }
                        x=aPar.x;
                        y=aPar.y;
                    }
                }
            }
            textPanel.append(str);
            //IJ.write(str);
        }
        //Now do the fancy stuff when requested:
        //Makes a new stack with objects labeled with track nr
        //optionally also displays centroid position
        if (bShowPositions)
        {
            int trackNr;
            ImageStack newstack=imp.createEmptyStack();
            int yWidth=newstack.getWidth();
            for (int i=0;i<nFrames;i++)
            {
                newstack.addSlice(stack.getSliceLabel(i+1),stack.getProcessor(i+1).crop());
                ImageProcessor nip=newstack.getProcessor(i+1);
                nip.setColor(Color.black);
                //hack to only show tracks longer than minTrackLength
                trackNr=0;
                for (int a=1;a<validTracks.length;a++)
                    if (validTracks[a]!=-1)//If it is a valid track...
                    {
                        trackNr++;
                        List bTrack=(ArrayList)theTracks.get(a-1);
                        if (i>=validTracks[a]&&i<validTracks[a]+bTrack.size())
                        {
                            particle aPar=(particle)bTrack.get(i-validTracks[a]);
                            str=""+trackNr+"="+(int)aPar.x+","+(int)aPar.y;
                            // we could do someboundary testing here to place the labels better when we are close to the edge
                            nip.moveTo((int)aPar.x+5,doOffset((int)aPar.y,yWidth,5));
                            //nip.moveTo(doOffset((int)aParticle.x,xHeight,5),doOffset((int)aParticle.y,yWidth,5));
                            nip.drawString(str);
                        }
                    }
                IJ.showProgress((double)i/nFrames);
            }
            ImagePlus nimp=new ImagePlus(imp.getTitle()+" labels",newstack);
            nimp.show();
        }
        int b=1;
        for (ListIterator iT=theTracks.listIterator();iT.hasNext();)
        {
            iT.next();
            if (validTracks[b++]==-1)
                iT.remove();//Remove the track if is shorter than the threshold
        }
        // total length traversed
        if (bShowPathLengths)
        {
            double[] lengths=new double[trackCount];
            int[] frames=new int[trackCount];
            int displayTrackNr=0;
            if (cc==-1)
                textPanel.append("Frame rate not included in the head file\n"+"Speed is calculated from the fps provided by the user");
//                IJ.write("Frame rate not included in the head file\n"+"Speed is calculated from the fps provided by the user");
            else
                textPanel.append("\nFrames per second\t"+info);
//                IJ.write("\nFrames per second\t"+info);
            //------------------------------------------------------------------
//            IJ.write("\nTrack \tLength\tSpeed\tNr of frames");
//            IJ.write("\nTrack \tLength\tNr of frames");
            textPanel.append("\nTrack \tLength\tNr of frames");
            for (ListIterator iT=theTracks.listIterator();iT.hasNext();)
            {
                List bTrack=(ArrayList)iT.next();
                displayTrackNr++;
                ListIterator jT=bTrack.listIterator();
                particle oldParticle=(particle)jT.next();
                frames[displayTrackNr-1]=bTrack.size();
                for (;jT.hasNext();)
                {
                    particle newParticle=(particle)jT.next();
                    lengths[displayTrackNr-1]+=Math.hypot(oldParticle.x-newParticle.x,oldParticle.y-newParticle.y);
                    oldParticle=newParticle;
                }
            }
            for (int i=0;i<displayTrackNr;i++)
//                IJ.write(i+1+"\t"+(float)lengths[i]+"\t"+(float)(fps*(float)lengths[i]/(float)frames[i])+"\t"+(int)frames[i]);
//                IJ.write(i+1+"\t"+(float)lengths[i]+"\t"+(int)frames[i]);
                textPanel.append(i+1+"\t"+(float)lengths[i]+"\t"+(int)frames[i]);
        }
        // 'map' of tracks
        if (bShowPaths)
        {
            if (imp.getCalibration().scaled())
            {
                IJ.showMessage("MultiTracker","Cannot display paths if image is spatially calibrated");
                return;
            }
            ImageProcessor ips=new ByteProcessor(imp.getWidth(),imp.getHeight());
            ips.setColor(Color.white);
            ips.fill();
            trackCount=0;
            int color;
            for (ListIterator iT=theTracks.listIterator();iT.hasNext();)
            {
                trackCount++;
                List bTrack=(ArrayList)iT.next();
                ListIterator jT=bTrack.listIterator();
                particle oldParticle=(particle)jT.next();
                for (;jT.hasNext();)
                {
                    particle newParticle=(particle)jT.next();
                    color=Math.min(trackCount+1,254);
                    ips.setValue(color);
                    ips.moveTo((int)oldParticle.x,(int)oldParticle.y);
                    ips.lineTo((int)newParticle.x,(int)newParticle.y);
                    oldParticle=newParticle;
                }
            }
            new ImagePlus("Paths",ips).show();
        }
    }

    // Utility functions
    int doOffset (int center,int maxSize,int displacement)
    {
        return ((center-displacement)<2*displacement)?(center+4*displacement):(center-displacement);
    }
}
