/**
 * Otsu_Limited.java
 * Created on 15 October 2009, 16:21
 *
 * Otsu thresholding algorithm, but allows to make an image black if the
 * threshold is too low or the image very dark.
 * This plugin do not create a new image.
 *
 * Copyright (c) 2009 by Manuel Guillermo Forero-Vargas
 * e-mail: mgforero@yahoo.es
 *
 * This plugin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this plugin; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;

public class Otsu_Limited implements PlugInFilter
{
    static int mingrey=100;//Minimum maximum grey level in an image
    static int mint=30;//Minimum threshold
    int oldt=255;

    public int setup(String arg, ImagePlus imp)
    {
        if (arg.equals("about"))
        {
            showAbout();
            return DONE;
        }
        //----------------------------------------------------------------------
        //Creates a dialog box, allowing the user to enter the requested
        //minimum maximum grey level and minimum threshold in an image.
        GenericDialog gd=new GenericDialog("Otsu Limited",IJ.getInstance());
        gd.addNumericField("Minimum_maximum_grey_level:",mingrey,0);
        gd.addNumericField("Minimum_threshold:",mint,0);
        gd.showDialog();
        if (gd.wasCanceled())
            return DONE;
        mingrey=(int)gd.getNextNumber();
        mint=(int)gd.getNextNumber();
        return IJ.setupDialog(imp,DOES_8G+DOES_8C);
    }

    public void run(ImageProcessor ip)
    {
        float[] normalizedHistogram;
        float n;//Etha
        float nOpt;//Optimum Etha
        float ut,uT;//Total mean
        float u0,u1;//Mean of clusters 1 and 2 (Background and object)
        float varT;//Total variance
        float w0,w1;
        int i,q;
        int threshold;
        int minGris,maxGris;//Limits of the histogram
     
        int size=ip.getHeight()*ip.getWidth();
        int[] histogram=((ByteProcessor) ip).getHistogram();
        normalizedHistogram=new float[256];
        //----------------------------------------------------------------------
        //Gets limits of the histogram
        //Returns the first intensity level that appears in the histogram
        minGris=0;
        while (histogram[minGris]==0)//Sweeps the histogram
            minGris++;
        //Returns the last intensity level that appears in the histogram
        maxGris=255;//Levels-1;
        while (histogram[maxGris]==0)
            maxGris--;
        if (maxGris<mingrey)
            threshold=oldt;
        else
        {
            for(i=minGris;i<=maxGris;i++)
                normalizedHistogram[i]=(float)histogram[i]/(float)size;
            //Calculates global variables
            for (q=minGris,uT=0;q<=maxGris;q++)//Calculates total mean
                uT+=normalizedHistogram[q]*(float)q;//q*P(q)
            for (q=minGris,varT=0;q<=maxGris;q++)//Calculates total variance
                varT+=((float)(q)-uT)*((float)(q)-uT)*normalizedHistogram[q];
            nOpt=0;
            threshold=minGris;
            w0=normalizedHistogram[minGris];
            ut=normalizedHistogram[minGris]*(float)minGris;
            //Calculates etha for each gray level
            for (q=minGris+1;q<=maxGris;q++)//Sweeps the histogram
            {
                w0+=normalizedHistogram[q];
                w1=1-w0;
                ut+=normalizedHistogram[q]*(float)(q);
                u0=(w0!=0)?ut/w0:400000000;//Makes this result huge
                u1=(w1!=0)?(uT-ut)/w1:400000000;//Same for u1
                if ((n=(w0*w1*(u0-u1)*(u0-u1))/varT)>nOpt)//Looks for the biggest etha
                {
                    nOpt=n;
                    threshold=q;
                }
            }
            if (threshold<mint)//Avoids very strong changes in the threshold from one image to another, when the image is too dark
                threshold=oldt;
            else
                oldt=threshold;
        }
        ip.threshold(threshold);
    }

    void showAbout()
    {
        IJ.showMessage("About Otsu_Limited...",
        "This plug-in filter calculates the Otsu Threshold of 8-bit images.");
    }
}
