/**
* Particle_Remover_Statistics.java
* Created on February 2005 by Manuel Guillermo Forero-Vargas
* e-mail: mgforero@yahoo.es
*
* This is an extended ParticleAnalyzer that creates a copy of the
* image that has each measured particle erased to the current foreground color.
* For stacks,the particles are erased in place.
*/
import ij.*;
import ij.measure.*;
import ij.plugin.*;
import ij.plugin.filter.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;

public class Particle_Remover_Statistics implements PlugIn,Measurements
{
    public void run(String arg)
    {
        ImagePlus imp=IJ.getImage();
        if (imp==null||imp.getType()!=ImagePlus.GRAY8)
        {
            IJ.error("8-bit grayscale image required");
            return;
        }
        int stackSize=imp.getStackSize();
        //----------------------------------------------------------------------
        //Defines the descriptors to be calculated in the analysis of the particles
        int measurements=Analyzer.getMeasurements();
        measurements|=AREA+PERIMETER+CIRCULARITY;
        Analyzer.setMeasurements(measurements);
        Analyzer.resetCounter();
        //----------------------------------------------------------------------
        HuParticleAnalyzer pa=new HuParticleAnalyzer();
        int flags=pa.setup("",imp);
        if (flags==PlugInFilter.DONE)
            return;
        if ((flags&PlugInFilter.DOES_STACKS)==0)
            stackSize=1;
        //----------------------------------------------------------------------
        for (int i=1;i<=stackSize;i++)
        {
            //----------------------------------------------------------------------
            if (IJ.escapePressed())
			{
                IJ.beep();
                return;
            }
            IJ.showProgress((double)i/stackSize);
            IJ.showStatus("a: "+i+"/"+stackSize);
            if (stackSize>1)
                imp.setSlice(i);
            ImageProcessor ip=imp.getProcessor();
            pa.run(imp,ip);
            if (pa.error)
                break;
            ip.setPixels(pa.ip2.getPixels());
        }
        IJ.showProgress(1.0);
            //----------------------------------------------------------------------
            //Method employed to stop the plugin if the window is closed
            ImageWindow win=imp.getWindow();
            if (win==null)
                return;
            //----------------------------------------------------------------------
        imp.updateAndDraw();
    }	
}

class HuParticleAnalyzer extends ParticleAnalyzer
{
	// Overrides method with the same in AnalyzeParticles that's called once for each particle
    boolean error;
    boolean firstTime=true;
//The following variables are made statics to keep the values the following time the window is opened
    private static double minCircularity=0;//Minimum circularity of the accepted objects
    private static double maxCircularity=1;//Maximum circularity of the accepted objects
    private static double minGeoExcentricity=0;//Minimum geometric excentricity of the accepted objects
    private static double maxGeoExcentricity=1;//Maximum geometric excentricity of the accepted objects
    private static double minImExcentricity=0;//Minimum image excentricity of the accepted objects
    private static double maxImExcentricity=1;//Maximum image excentricity of the accepted objects
    private static double minPhi0=0,maxPhi0=1,minPhi1=0,maxPhi1=1,minPhi2=0,maxPhi2=1,minPhi7=0,maxPhi7=1;
    private static int minArea=33,maxArea=500;//Maximum and minimum area in pixels of the accepted objects
    int[] Descriptor=new int[11];
    double[] Phi=new double[8];
    ImageProcessor ip2;

    public boolean setfilters()
    {
        //----------------------------------------------------------------------
        /*Creates a dialog box,allowing the user to enter the requested
        maximum and minimum area of the objects to be retained.*/
        GenericDialog gda=new GenericDialog("Particle remover");//getInstance());
        gda.addNumericField("Minimum_area",minArea,0,2,"pixels");
        gda.addNumericField("Maximum_area",maxArea,0,4,"pixels");
        gda.addNumericField("Minimum_circularity",minCircularity,2);
        gda.addNumericField("Maximum_circularity",maxCircularity,2);
        gda.addNumericField("Minimum_geometric excentrity",minGeoExcentricity,2);
        gda.addNumericField("Maximum_geometric excentrity",maxGeoExcentricity,2);
        gda.addNumericField("Minimum_image excentrity",minImExcentricity,2);
        gda.addNumericField("Maximum_image excentrity",maxImExcentricity,2);
        gda.addNumericField("Minimum_1st Hu�s_moment",minPhi0,5);
        gda.addNumericField("Maximum_1st Hu�s_moment",maxPhi0,5);
        gda.addNumericField("Minimum_2nd Hu�s moment",minPhi1,5);
        gda.addNumericField("Maximum_2nd Hu�s moment",maxPhi1,5);
        gda.addNumericField("Minimum_3rd Hu�s moment",minPhi2,5);
        gda.addNumericField("Maximum_3rd Hu�s moment",maxPhi2,5);
        gda.addNumericField("Minimum_11th Hu�s moment",minPhi7,5);
        gda.addNumericField("Maximum_11th Hu�s moment",maxPhi7,5);
        gda.showDialog();
        if (gda.wasCanceled())
            return false;
        minArea=(int)gda.getNextNumber();
        maxArea=(int)gda.getNextNumber();
        minCircularity=(double)gda.getNextNumber();
        maxCircularity=(double)gda.getNextNumber();
        minGeoExcentricity=(double)gda.getNextNumber();
        maxGeoExcentricity=(double)gda.getNextNumber();
        minImExcentricity=(double)gda.getNextNumber();
        maxImExcentricity=(double)gda.getNextNumber();
        minPhi0=(double)gda.getNextNumber();
        maxPhi0=(double)gda.getNextNumber();
        minPhi1=(double)gda.getNextNumber();
        maxPhi1=(double)gda.getNextNumber();
        minPhi2=(double)gda.getNextNumber();
        maxPhi2=(double)gda.getNextNumber();
        minPhi7=(double)gda.getNextNumber();
        maxPhi7=(double)gda.getNextNumber();
        //----------------------------------------------------------------------
        return true;
    }
    
    public void run(ImagePlus imp,ImageProcessor ip)
    {
        ip2=ip.duplicate();//Make a copy of this image
        ip2.setColor(255);//Color used to fill the objects to be eliminated
        slice++;
        error=!analyze(imp,ip);
    }
    
    //Overrides method with the same in AnalyzeParticles that's called once for each particle
    protected void saveResults(ImageStatistics stats,Roi roi)
    {
        double circularity;
        double geoExcentricity;
        double imExcentricity;
        
        analyzer.saveResults(stats,roi);
        if (firstTime)
            init();
        Phi=HuFlusser(roi);
        rt.addValue(Descriptor[0],Phi[0]);
        rt.addValue(Descriptor[1],Phi[1]);
        rt.addValue(Descriptor[2],Phi[2]);
        rt.addValue(Descriptor[3],Phi[3]);
        rt.addValue(Descriptor[4],Phi[4]);
        rt.addValue(Descriptor[5],Phi[5]);
        rt.addValue(Descriptor[6],Phi[6]);
        rt.addValue(Descriptor[7],Phi[7]);
        double sqphi1=Math.sqrt(Phi[1]);
        geoExcentricity=Math.sqrt(2.0*sqphi1/(Phi[0]+sqphi1));
        rt.addValue(Descriptor[8],geoExcentricity);
        imExcentricity=Math.sqrt((Phi[0]-sqphi1)/(Phi[0]+sqphi1));
        rt.addValue(Descriptor[9],imExcentricity);
        rt.addValue(Descriptor[10],stats.pixelCount);//Adds the area measured in pixels
        if (showResults)
            analyzer.displayResults();
        int region=rt.getCounter()-1;
        circularity=rt.getValue(ResultsTable.CIRCULARITY,region);
        if (circularity==-1&minCircularity>=0.5&maxCircularity==1)
            circularity=1;
        if (stats.pixelCount<minArea|stats.pixelCount>maxArea|
        circularity<minCircularity|circularity>maxCircularity|
        geoExcentricity<minGeoExcentricity|geoExcentricity>maxGeoExcentricity|
        imExcentricity<minImExcentricity|imExcentricity>maxImExcentricity|
        Phi[0]<minPhi0|Phi[0]>maxPhi0|Phi[1]<minPhi1|Phi[1]>maxPhi1|
        Phi[2]<minPhi2|Phi[2]>maxPhi2|Phi[7]<minPhi7|Phi[7]>maxPhi7)//Condition for elimination
        {
            ip2.setRoi(roi.getBoundingRect());
            ip2.fill(roi.getMask());//Delete the roi in the image
        }
    }

    // Overrides method with the same in AnalyzeParticles that dsplays a modal options dialog.
    public boolean showDialog()
    {
        if (!super.showDialog())
            return false;
        return setfilters()?true:false;
    }
    
    void init()
    {
        Descriptor[0]=getColumnID("Phi0");
        Descriptor[1]=getColumnID("Phi1");
        Descriptor[2]=getColumnID("Phi2");
        Descriptor[3]=getColumnID("Phi3");
        Descriptor[4]=getColumnID("Phi4");
        Descriptor[5]=getColumnID("Phi5");
        Descriptor[6]=getColumnID("Phi6");
        Descriptor[7]=getColumnID("Phi11");
        Descriptor[8]=getColumnID("Geometric Excentricity");
        Descriptor[9]=getColumnID("Image Excentricity");
        Descriptor[10]=getColumnID("Pixels");
        firstTime=false;
    }
	
    int getColumnID(String name)
    {
        int id=rt.getFreeColumn(name);
        if (id==ResultsTable.COLUMN_IN_USE)
            id=rt.getColumnIndex(name);
        return id;
    }

    //Receive the roi,where the pixels belonging to the object has a value of 255 and the background is 0.
    double[] HuFlusser(Roi roi)
    {
        double      in,jn;
        int         offset,order=4;
        int         r,s;
        int         maxorder=order+1;
        double      centroidI,centroidJ;
        double[][]  m=new double[maxorder][maxorder];
        double[][]  mu=new double[maxorder][maxorder];
        double[][]  n=new double[maxorder][maxorder];
        double[]    phi=new double[8];
        //----------------------------------------------------------------
        ImageProcessor ip=roi.getMask();
        int         width=ip.getWidth();
        Rectangle   re=ip.getRoi();
        byte[]      pixels=(byte[])ip.getPixels();

        //----------------------------------------------------------------
        //Discrete general moments
        for (int i=re.y;i<re.y+re.height;i++)
        {
            offset=i*width;
            for (int j=re.x;j<re.x+re.width;j++)
                if ((pixels[offset+j]&0xff)==255)
                    for (r=0;r<=order;r++)
                        for (s=0;s<=order;s++)
                            m[r][s]+=Math.pow(i,r)*Math.pow(j,s);
        }
        //----------------------------------------------------------------
        //Calcul of centroids
        if (m[0][0]==0)
        {
            centroidI=0;
            centroidJ=0;
        }
        else
        {
            centroidI=m[1][0]/m[0][0];
            centroidJ=m[0][1]/m[0][0];
        }
        //----------------------------------------------------------------
        //Calcul of moments invariant to traslation
        for (int i=re.y;i<re.y+re.height;i++)
        {
            offset=i*width;
            in=i-centroidI;
            for (int j=re.x;j<re.x+re.width;j++)
                if ((pixels[offset+j]&0xff)==255)
                {
                    jn=j-centroidJ;
                    for (r=0;r<=order;r++)
                        for (s=0;s<=order;s++)
                            mu[r][s]+=Math.pow(in,r)*Math.pow(jn,s);
                }
        }
        //----------------------------------------------------------------
        //Calcul of moments invariants to traslation and change of scale (etha)
       for (r=0;r<=order;r++)
            for (s=0;s<=order;s++)
                if (r+s>=2)
                    n[r][s]=(mu[0][0]==0)?0:mu[r][s]/(Math.pow(mu[0][0],(r+s+2.0)/2.0));
        //----------------------------------------------------------------
        //Calcul of the Hu moments invariants to traslation,rotation y change of scale
        phi[0]=n[2][0]+n[0][2];
        phi[1]=(n[2][0]-n[0][2])*(n[2][0]-n[0][2])+4*n[1][1]*n[1][1];
        phi[2]=(n[3][0]-3*n[1][2])*(n[3][0]-3*n[1][2])+(3*n[2][1]-n[0][3])*(3*n[2][1]-n[0][3]);
        phi[3]=(n[3][0]+n[1][2])*(n[3][0]+n[1][2])+(n[2][1]+n[0][3])*(n[2][1]+n[0][3]);
        phi[4]=(n[3][0]-3*n[1][2])*(n[3][0]+n[1][2])*((n[3][0]+n[1][2])*(n[3][0]+n[1][2])-3*(n[2][1]+n[0][3])*(n[2][1]+n[0][3]))+(3*n[2][1]-n[0][3])*(n[2][1]+n[0][3])*(3*(n[3][0]+n[1][2])*(n[3][0]+n[1][2])-(n[2][1]+n[0][3])*(n[2][1]+n[0][3]));
        phi[5]=(n[2][0]-n[0][2])*((n[3][0]+n[1][2])*(n[3][0]+n[1][2])-(n[2][1]+n[0][3])*(n[2][1]+n[0][3]))+4*n[1][1]*(n[3][0]+n[1][2])*(n[2][1]+n[0][3]);
        phi[6]=(3*n[2][1]-n[0][3])*(n[3][0]+n[1][2])*((n[3][0]+n[1][2])*(n[3][0]+n[1][2])-3*(n[2][1]+n[0][3])*(n[2][1]+n[0][3]))+(3*n[1][2]-n[3][0])*(n[2][1]+n[0][3])*(3*(n[3][0]+n[1][2])*(n[3][0]+n[1][2])-(n[2][1]+n[0][3])*(n[2][1]+n[0][3]));
        //----------------------------------------------------------------
        //Calcul of moment invariant to traslatio,rotation and change of scale
        phi[7]=n[4][0]+2*n[2][2]+n[0][4];
        return phi;
    }
}
